from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel

from .base import Channel, RetailerProgramData, Tag
from .products import BaseProductModel
from .shipping import Recipient, Shipment, ShippingTerms


class OrderStatus(Enum):
    PROCESSING = "Processing"
    EXCEPTION = "Exception"
    PARTIALLY_FULFILLED = "PartiallyFulfilled"
    FULFILLED = "Fulfilled"
    CANCELLED = "Cancelled"
    IMPORT_REVIEW = "ImportReview"


class OrderType(Enum):
    DTC = "DTC"
    DROPSHIP = "DropShip"
    B2B = "B2B"


class OrderedProduct(BaseProductModel):
    quantity: int
    quantity_unit_of_measure_code: str
    gtin: str
    upc: str
    unit_price: Decimal
    external_line_id: int


class Order(BaseModel):
    id: int
    created_date: datetime
    purchase_date: Optional[datetime]
    reference_id: Optional[str]
    order_number: Optional[str]
    status: OrderStatus
    type: OrderType
    channel: Optional[Channel]
    shipping_method: Optional[str]
    recipient: Optional[Recipient]
    products: List[OrderedProduct]
    tags: List[Tag]
    shipments: List[Shipment]
    gift_message: Optional[str]
    shipping_terms: ShippingTerms
    retailer_program_data: RetailerProgramData


__all__ = ["Order", "OrderStatus", "OrderType"]



from datetime import datetime
from decimal import Decimal
from typing import List, Optional

from pydantic import BaseModel

from .base import Channel
from .inventory import InventoryItemSummary


class FulfillableItemByFulfillmentCenter(BaseModel):
    id: int
    name: str
    fulfillable_quantity: int
    onhand_quantity: int
    committed_quantity: int


class BaseProductModel(BaseModel):
    id: int
    reference_id: Optional[str]
    sku: Optional[str]


class BundleRootInformation(BaseModel):
    id: int
    name: str


class Product(BaseProductModel):
    name: str
    bundle_root_information: Optional[BundleRootInformation]
    created_date: datetime
    channel: Channel
    quantity: Optional[int]
    gtin: Optional[str]
    upc: Optional[str]
    unit_price: Optional[Decimal]
    barcode: Optional[str]

    total_fulfillable_quantity: int
    total_onhand_quantity: int
    total_committed_quantity: int
    fulfillable_inventory_items: List[InventoryItemSummary]
    fulfillable_quantity_by_fulfillment_center: List[FulfillableItemByFulfillmentCenter]


__all__ = ["Product"]



from decimal import Decimal
from enum import Enum
from typing import Optional

from pydantic import BaseModel


class RetailerProgramDataType(Enum):
    MARK_FOR = "MarkFor"
    SHIP_FROM = "ShipFrom"


class Channel(BaseModel):
    id: int
    name: str


class Address(BaseModel):
    address1: Optional[str]
    address2: Optional[str]
    city: Optional[str]
    state: Optional[str]
    country: Optional[str]
    zip_code: Optional[str]


class RetailerProgramData(Address):
    type: Optional[RetailerProgramDataType]
    company_name: Optional[str]


class BaseMeasurementModel(BaseModel):
    total_weight_oz: Decimal
    length_in: Decimal
    width_in: Decimal
    depth_in: Decimal


class Tag(BaseModel):
    name: str
    value: str



import logging
import os
import time
from datetime import datetime, timedelta
from typing import Iterator, List

from pydantic import ValidationError
from requests import Session
from requests.exceptions import HTTPError

from .models.inventory import InventoryItem
from .models.orders import Order
from .models.products import Product
from .models.shipping import Shipment, ShipmentEvent

logger = logging.getLogger(__name__)


class ShipBobClient(Session):
    BASE_URL = "https://api.shipbob.com/1.0"

    ACCESS_TOKEN = os.getenv("SHIPBOB_ACCESS_TOKEN")

    MAX_PAGE_SIZE: int = 250
    MAX_REQUESTS_PER_MINUTE: int = 150  # As per ShipBob docs, 150 reqs/min sliding window

    def __init__(self):
        super().__init__()
        if not self.ACCESS_TOKEN:
            raise ValueError("Please set SHIPBOB_ACCESS_TOKEN environment variable")

        self.headers.update(
            {"Content-Type": "application/json", "Authorization": f"Bearer {self.ACCESS_TOKEN}"}
        )

        self._request_bucket = []

    def _clear_request_bucket(self):
        now = datetime.now()
        one_minute_old = now - timedelta(seconds=60)
        self._request_bucket = [rt for rt in self._request_bucket if rt >= one_minute_old]

    def request(self, method, url, retry_on_rate_limit=True, *args, **kwargs):
        while len(self._request_bucket) >= self.MAX_REQUESTS_PER_MINUTE:
            logger.warning("Close to rate limit. Sleeping for one second")
            self._clear_request_bucket()
            time.sleep(1)

        self._request_bucket.append(datetime.now())

        full_url = "/".join((self.BASE_URL, url.lstrip("/")))
        response = super().request(method, full_url, *args, **kwargs)
        try:
            response.raise_for_status()
        except HTTPError as e:
            if e.response.status_code == 429 and retry_on_rate_limit:
                logger.warning("We hit the rate limit.")
                try:
                    retry_after: int = int(e.response.headers.get("Retry-After", 5))
                except ValueError:
                    retry_after: int = 10
                logger.warning(f"Waiting {retry_after} seconds to try again...")
                time.sleep(retry_after)
                return self.request(method, url, retry_on_rate_limit=False, *args, **kwargs)
            else:
                raise e
        return response

    def get_orders(self, **params) -> List[Order]:
        response = self.request("GET", "order", params=params)
        response.raise_for_status()
        return [Order(**order_data) for order_data in response.json()]

    def get_order_shipments(self, order: Order, **params) -> List[Shipment]:
        response = self.request("GET", f"order/{order.id}/shipment", params=params)
        response.raise_for_status()
        return [Shipment(**shipment_data) for shipment_data in response.json()]

    def get_shipment_events(self, shipment: Shipment, **params) -> List[ShipmentEvent]:
        try:
            response = self.request("GET", f"shipment/{shipment.id}/timeline", params=params)
            response.raise_for_status()
            return [ShipmentEvent(**timeline) for timeline in response.json()]
        except ValidationError as exc:
            logger.warn(f"Failed to get events for {shipment}: {exc}")
            return []

    def get_products(self, **params) -> List[Product]:
        response = self.request("GET", "product", params=params)
        response.raise_for_status()
        return [Product(**product_data) for product_data in response.json()]

    def get_inventory_items(self, **params) -> List[InventoryItem]:
        response = self.request("GET", "inventory", params=params)
        response.raise_for_status()
        return [InventoryItem(**item_data) for item_data in response.json()]

    def iterate_products(self, **params) -> Iterator[Product]:
        params.pop("page", None)
        params.pop("limit", None)
        params["SortOrder"] = "Oldest"

        page = 1
        while products := self.get_products(page=page, limit=self.MAX_PAGE_SIZE, **params):
            logger.debug(f"Getting page {page} of products")
            for product in products:
                yield product
            page += 1

    def iterate_orders(self, **params) -> Iterator[Order]:
        params.pop("page", None)
        params.pop("limit", None)

        page = 1
        while orders := self.get_orders(page=page, limit=self.MAX_PAGE_SIZE, **params):
            logger.debug(f"Getting page {page} of orders")
            for order in orders:
                yield order
            page += 1

    def iterate_inventory_items(self, **params) -> Iterator[InventoryItem]:
        params.pop("page", None)
        params.pop("limit", None)

        page = 1
        while inventory_items := self.get_inventory_items(
            page=page, limit=self.MAX_PAGE_SIZE, **params
        ):
            logger.debug(f"Getting page {page} of inventory_items")
            for inventory_item in inventory_items:
                yield inventory_item
            page += 1



from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Dict, List, Optional

from pydantic import BaseModel

from .base import Address, BaseMeasurementModel
from .products import BaseProductModel


class CarrierType(Enum):
    PARCEL = "Parcel"
    FREIGHT = "Freight"


class PaymentTermType(Enum):
    COLLECT = "Collect"
    THIRD_PARTY = "ThirdParty"
    PREPAID = "Prepaid"
    MERCHANT_RESPONSIBLE = "MerchantResponsible"


class ShipmentStatus(Enum):
    NONE = None
    PROCESSING = "Processing"
    PENDING = "Pending"
    COMPLETED = "Completed"
    EXCEPTION = "Exception"
    HELD = "OnHold"
    CANCELLED = "Cancelled"
    CLEAN_SWEEPED = "CleanSweeped"
    LABELED = "LabeledCreated"
    IMPORT_REVIEW = "ImportReview"


class PackageMaterialType(Enum):
    UNKNOWN = "Unknown"
    BOX = "Box"
    BUBBLE_MAILER = "BubbleMailer"
    POLYMAILER = "PolyMailer"
    FRAGILE_BOX = "FragileBox"
    POSTER_TUBE = "PosterTube"
    CUSTOM = "Custom"
    BOOKFOLD = "Bookfold"
    OWN_CONTAINER = "OwnContainer"
    UNDEFINED = "Undefined"


class EstimatedFulfillmentStatus(Enum):
    AWAITING_INVENTORY_ALLOCATION = "AwaitingInventoryAllocation"
    AWAITING_RESET = "AwaitingReset"
    UNAVAILABLE = "Unavailable"
    PENDING_ON_TIME = "PendingOnTime"
    FULFILLED_ON_TIME = "FulfilledOnTime"
    PENDING_LATE = "PendingLate"
    FULFILLED_LATE = "FulfilledLate"


class CartonBaseModel(BaseModel):
    type: str
    barcode: str
    measurements: BaseMeasurementModel


class Recipient(BaseModel):
    name: str
    address: Address
    email: Optional[str]
    phone_number: Optional[str]


class ShipmentStatusDetail(BaseModel):
    name: str
    description: str
    id: int
    inventory_id: Optional[int]
    exception_fullfilment_center_id: Optional[int]


class TrackingDetail(BaseModel):
    carrier: str
    tracking_number: str
    carrier_service: Optional[str]
    tracking_url: Optional[str]
    bol: Optional[str]
    shipping_date: Optional[datetime]
    pro_number: Optional[str]
    scac: Optional[str]


class ShippedInventoryItem(BaseModel):
    id: int
    name: str
    quantity: int
    quantity_committed: int
    lot: Optional[str]
    expiration_date: Optional[datetime]
    serial_numbers: List[str]
    is_dangerous_goods: bool


class ShippedProduct(BaseProductModel):
    name: str
    inventory_items: List[ShippedInventoryItem]


class Carton(CartonBaseModel):
    carton_details: List[ShippedProduct]


class ParentCarton(CartonBaseModel):
    cartons: List[Carton]


class FulfillmentCenter(BaseModel):
    id: int
    name: str


class Shipment(BaseModel):
    id: int
    store_order_id: Optional[str]
    order_id: int
    reference_id: Optional[str]
    created_date: datetime
    last_update_at: Optional[datetime]
    status: ShipmentStatus
    status_detail: Optional[ShipmentStatusDetail]
    location: Optional[FulfillmentCenter]
    invoice_amount: Decimal
    insurance_value: Optional[Decimal]
    ship_option: str
    package_material_type: PackageMaterialType
    tracking: Optional[TrackingDetail]
    products: List[ShippedProduct]
    parent_cartons: List[ParentCarton]
    measurements: Optional[BaseMeasurementModel]
    require_signature: bool
    estimated_fulfillment_date: Optional[datetime]
    estimated_fulfillment_date_status: EstimatedFulfillmentStatus
    actual_fulfllment_date: Optional[datetime]
    is_tracking_uploaded: bool
    gift_message: Optional[str]


class ShippingTerms(BaseModel):
    carrier_type: Optional[CarrierType]
    payment_term: Optional[PaymentTermType]


class ShipmentEvent(BaseModel):
    log_type_id: int
    log_type_name: str
    log_type_text: str
    timestamp: datetime
    metadata: Optional[Dict]


__all__ = [
    "CarrierType",
    "PaymentTermType",
    "ShipmentStatus",
    "PackageMaterialType",
    "EstimatedFulfillmentStatus",
    "FulfillmentCenter",
    "Recipient",
    "ShipmentStatusDetail",
    "TrackingDetail",
    "ShippedInventoryItem",
    "ShippedProduct",
    "Carton",
    "ParentCarton",
    "Shipment",
    "ShippingTerms",
    "ShipmentEvent",
]



from datetime import datetime
from decimal import Decimal
from typing import Any, List, Optional

from pydantic import BaseModel


class InventoryItemDimensions(BaseModel):
    weight: Decimal
    length: Decimal
    width: Decimal
    depth: Decimal


class FulfillableQuantityByFulfillmentCenter(BaseModel):
    id: int
    name: str
    fulfillable_quantity: int
    onhand_quantity: int
    committed_quantity: int
    awaiting_quantity: int
    internal_transfer_quantity: int


class FulfillableQuantityByLot(BaseModel):
    lot_number: Optional[int]
    expiration_date: Optional[datetime]
    fulfillable_quantity: int
    onhand_quantity: int
    committed_quantity: int
    awaiting_quantity: int
    internal_transfer_quantity: int
    fulfillable_quantity_by_fulfillment_center: List[FulfillableQuantityByFulfillmentCenter]


class InventoryItem(BaseModel):
    id: int
    name: str
    is_digital: bool
    is_case_pick: bool
    is_lot: bool
    dimensions: InventoryItemDimensions
    total_fulfillable_quantity: int
    total_onhand_quantity: int
    total_committed_quantity: int
    total_sellable_quantity: int
    total_awaiting_quantity: int
    total_exception_quantity: int
    total_internal_transfer_quantity: int
    total_backordered_quantity: int
    is_active: bool
    fulfillable_quantity_by_fulfillment_center: List[FulfillableQuantityByFulfillmentCenter]
    fulfillable_quantity_by_lot: List[FulfillableQuantityByLot]
    packaging_attribute: Optional[Any]


class InventoryItemSummary(BaseModel):
    id: int
    name: str
    quantity: int


__all__ = ["InventoryItem", "InventoryItemSummary"]
